
import theano
import numpy
import sys
import os

import _setup_test_env  # noqa
from returnn.theano.layers.base import Container, SourceLayer
from returnn.theano.layers.rec import RecurrentUnitLayer

Container.initialize_rng()

# TODO more sane data
index = theano.shared(numpy.array([[1]]), name="i")
source = SourceLayer(n_out=2, x_out=theano.shared(numpy.array([[[1.0, -2.0]]], dtype='float32'), name="x", index=index))


def test_RecurrentUnitLayer_init():
  RecurrentUnitLayer(n_out=3, sources=[source], index=index)

def test_RecurrentUnitLayer_init_sampling2():
  RecurrentUnitLayer(n_out=3, sources=[source], index=index, sampling=2)



#!/usr/bin/env python3

"""
Benchmarking LSTM training times for different LSTM implementations, both on GPU and CPU.

Some runtimes:

GeForce GTX 680:
  Settings:
  {'batch_size': 2000,
   'chunking': '50:25',
   'max_seqs': 40,
   'n_hidden': 500,
   'num_layers': 5,
   'num_seqs': 500}
  Final results:
    GPU:CudnnLSTM: 0:00:12.3727
    GPU:NativeLSTM: 0:00:19.0306
    GPU:LSTMBlockFused: 0:00:41.5129
    GPU:LSTMBlock: 0:00:52.6296
    GPU:StandardLSTM: 0:00:53.2629
    GPU:BasicLSTM: 0:00:53.6150
    CPU:NativeLSTM: 0:02:37.1753
    CPU:LSTMBlockFused: 0:03:17.5448
    CPU:BasicLSTM: 0:03:28.9882
    CPU:StandardLSTM: 0:03:29.3730
    CPU:LSTMBlock: 0:03:33.7039

  Settings:
  {'batch_size': 2000,
   'chunking': '0',
   'max_seqs': 50,
   'n_hidden': 500,
   'num_layers': 5,
   'num_seqs': 500}
  Final results:
    GPU:CudnnLSTM: 0:00:12.8325
    GPU:NativeLSTM: 0:00:15.8518
    GPU:LSTMBlockFused: 0:01:05.2157
    GPU:StandardLSTM: 0:01:20.6564
    GPU:BasicLSTM: 0:01:20.9896
    GPU:LSTMBlock: 0:01:21.4046
    CPU:NativeLSTM: 0:02:12.5029
    CPU:LSTMBlockFused: 0:03:39.0270
    CPU:LSTMBlock: 0:03:51.9667
    CPU:StandardLSTM: 0:03:56.6404
    CPU:BasicLSTM: 0:03:58.1545
"""

from __future__ import print_function
import sys
import time
from argparse import ArgumentParser
from pprint import pprint

import _setup_returnn_env  # noqa
from returnn.util import better_exchook
from returnn.log import log
from returnn.config import Config
from returnn.util.basic import hms_fraction, describe_returnn_version, describe_tensorflow_version
from returnn.tf.engine import Engine
from returnn.tf.util.basic import setup_tf_thread_pools, is_gpu_available, print_available_devices
from returnn.datasets import init_dataset, Dataset


LstmCellTypes = [
  "BasicLSTM", "StandardLSTM",
  "LSTMBlock", "LSTMBlockFused",
  "NativeLSTM", "NativeLstm2", "NativeLstmLowMem",
  "CudnnLSTM"
]

GpuOnlyCellTypes = ["CudnnLSTM"]

# Fixed by dataset. See make_config_dict().
_input_dim = 9
_output_dim = 2

# You can play around with these. E.g. use "chunking=0", "max_seqs=20" as command-line args.
base_settings = {
  "num_layers": 5,  # number of LSTM layers
  "n_hidden": 500,  # per direction, per layer, the number of units
  "num_seqs": 500,  # for the dataset generation
  "batch_size": 2000,  # upper limit for n_seqs * max_seq_len in a batch
  "max_seqs": 40,  # upper limit for n_seqs in a batch
  "chunking": "50:25"  # set to "0" to disable
}


def make_config_dict(lstm_unit, use_gpu):
  """
  :param str lstm_unit: "NativeLSTM", "LSTMBlock", "LSTMBlockFused", "CudnnLSTM", etc, one of LstmCellTypes
  :param bool use_gpu:
  :return: config dict
  :rtype: dict[str]
  """
  num_layers = base_settings["num_layers"]
  network = {}
  for i in range(num_layers):
    for direction in [-1, 1]:
      dir_str = {-1: "bwd", 1: "fwd"}[direction]
      layer = {"class": "rec", "unit": lstm_unit, "n_out": base_settings["n_hidden"], "direction": direction}
      if i > 0:
        layer["from"] = ["lstm%i_fwd" % i, "lstm%i_bwd" % i]
      network["lstm%i_%s" % (i + 1, dir_str)] = layer
  network["output"] = {
    "class": "softmax", "loss": "ce", "target": "classes",
    "from": ["lstm%i_fwd" % num_layers, "lstm%i_bwd" % num_layers]}
  return {
    "device": "gpu" if use_gpu else "cpu",
    "train": {"class": "Task12AXDataset", "num_seqs": base_settings["num_seqs"]},
    "num_inputs": _input_dim,
    "num_outputs": _output_dim,
    "num_epochs": 1,
    "model": None,  # don't save
    "tf_log_dir": None,  # no TF logs
    "tf_log_memory_usage": True,
    "network": network,
    # batching
    "batch_size": base_settings["batch_size"],
    "max_seqs": base_settings["max_seqs"],
    "chunking": base_settings["chunking"],
    # optimization
    "adam": True,
    "learning_rate": 0.01}


def benchmark(lstm_unit, use_gpu):
  """
  :param str lstm_unit: e.g. "LSTMBlock", one of LstmCellTypes
  :param bool use_gpu:
  :return: runtime in seconds of the training itself, excluding initialization
  :rtype: float
  """
  device = {True: "GPU", False: "CPU"}[use_gpu]
  key = "%s:%s" % (device, lstm_unit)
  print(">>> Start benchmark for %s." % key)
  config = Config()
  config.update(make_config_dict(lstm_unit=lstm_unit, use_gpu=use_gpu))
  dataset_kwargs = config.typed_value("train")
  Dataset.kwargs_update_from_config(config, dataset_kwargs)
  dataset = init_dataset(dataset_kwargs)
  engine = Engine(config=config)
  engine.init_train_from_config(config=config, train_data=dataset)
  print(">>> Start training now for %s." % key)
  start_time = time.time()
  engine.train()
  runtime = time.time() - start_time
  print(">>> Runtime of %s: %s" % (key, hms_fraction(runtime)))
  engine.finalize()
  return runtime


def main():
  """
  Main entry.
  """
  global LstmCellTypes
  print("Benchmarking LSTMs.")
  better_exchook.install()
  print("Args:", " ".join(sys.argv))
  arg_parser = ArgumentParser()
  arg_parser.add_argument("cfg", nargs="*", help="opt=value, opt in %r" % sorted(base_settings.keys()))
  arg_parser.add_argument("--no-cpu", action="store_true")
  arg_parser.add_argument("--no-gpu", action="store_true")
  arg_parser.add_argument("--selected", help="comma-separated list from %r" % LstmCellTypes)
  arg_parser.add_argument("--no-setup-tf-thread-pools", action="store_true")
  args = arg_parser.parse_args()
  for opt in args.cfg:
    key, value = opt.split("=", 1)
    assert key in base_settings
    value_type = type(base_settings[key])
    base_settings[key] = value_type(value)
  print("Settings:")
  pprint(base_settings)

  log.initialize(verbosity=[4])
  print("Returnn:", describe_returnn_version(), file=log.v3)
  print("TensorFlow:", describe_tensorflow_version(), file=log.v3)
  print("Python:", sys.version.replace("\n", ""), sys.platform)
  if not args.no_setup_tf_thread_pools:
    setup_tf_thread_pools(log_file=log.v2)
  else:
    print("Not setting up the TF thread pools. Will be done automatically by TF to number of CPU cores.")
  if args.no_gpu:
    print("GPU will not be used.")
  else:
    print("GPU available: %r" % is_gpu_available())
  print_available_devices()

  if args.selected:
    LstmCellTypes = args.selected.split(",")
  benchmarks = {}
  if not args.no_gpu and is_gpu_available():
    for lstm_unit in LstmCellTypes:
      benchmarks["GPU:" + lstm_unit] = benchmark(lstm_unit=lstm_unit, use_gpu=True)
  if not args.no_cpu:
    for lstm_unit in LstmCellTypes:
      if lstm_unit in GpuOnlyCellTypes:
        continue
      benchmarks["CPU:" + lstm_unit] = benchmark(lstm_unit=lstm_unit, use_gpu=False)

  print("-" * 20)
  print("Settings:")
  pprint(base_settings)
  print("Final results:")
  for t, lstm_unit in sorted([(t, lstm_unit) for (lstm_unit, t) in sorted(benchmarks.items())]):
    print("  %s: %s" % (lstm_unit, hms_fraction(t)))
  print("Done.")


if __name__ == "__main__":
  main()



import tensorflow as tf
import returnn.tf.compat as tf_compat
from returnn.tf.layers.basic import LayerBase, _ConcatInputLayer, Loss, get_concat_sources_data_template
from returnn.tf.layers.rec import RecLayer
from returnn.tf.util.basic import Data, sparse_labels_with_seq_lens
from returnn.util.basic import softmax


class NeuralTransducerLayer(_ConcatInputLayer):
    """
    Creates a neural transducer based on the paper "A Neural Transducer": https://arxiv.org/abs/1511.04868.
    NOTE: Requires that the loss be neural_transducer_loss.
    NOTE: When training with BiLSTM as input, set an appropriate gradient clipping parameter.
    """
    layer_class = "neural_transducer"

    def __init__(self, transducer_hidden_units, n_out, transducer_max_width, input_block_size,
                 embedding_size, e_symbol_index, use_prev_state_as_start=False, **kwargs):
        """
        Initialize the Neural Transducer.
        :param int transducer_hidden_units: Amount of units the transducer should have.
        :param int n_out: The size of the output layer, i.e. the size of the vocabulary including <E> symbol.
        :param int transducer_max_width: The max amount of outputs in one NT block (including the final <E> symbol)
        :param int input_block_size: Amount of inputs to use for each NT block.
        :param int embedding_size: Embedding dimension size.
        :param int e_symbol_index: Index of e symbol that is used in the NT block. 0 <= e_symbol_index < num_outputs
        :param bool use_prev_state_as_start: Whether to use the last state of the previous recurrent layer as the ]
        initial state of the transducer. NOTE: For this to work, you have to watch out for:
        previous_layer.hidden_units = previous_layer.n_out = transducer.transducer_hidden_units
        """

        super(NeuralTransducerLayer, self).__init__(**kwargs)

        # TODO: Build optimized version

        # Get embedding
        from returnn.tf.util.basic import get_initializer
        initializer = get_initializer('glorot_uniform',
                                      seed=self.network.random.randint(2 ** 31),
                                      eval_local_ns={"layer": self})
        embeddings = self.add_param(tf_compat.v1.get_variable(shape=[n_out, embedding_size], dtype=tf.float32,
                                                              initializer=initializer, name='nt_embedding'),
                                    trainable=True, saveable=True)

        # Ensure encoder is time major
        encoder_outputs = self.input_data.get_placeholder_as_time_major()

        # Pad encoder outputs with zeros so that it is divisible by the input_block_size
        batch_size = tf.shape(encoder_outputs)[1]
        time_length_to_append = input_block_size - tf.mod(tf.shape(encoder_outputs)[0], input_block_size)
        padding_tensor = tf.zeros([time_length_to_append, batch_size, tf.shape(encoder_outputs)[2]],
                                  dtype=tf.float32)
        encoder_outputs = tf.concat([encoder_outputs, padding_tensor], axis=0)
        # Do assertions
        assert 0 <= e_symbol_index < n_out, 'NT: E symbol outside possible outputs!'

        # Get prev state as start state
        last_hidden = None
        if use_prev_state_as_start is True and isinstance(self.sources[0], RecLayer) is True:
            # TODO: add better checking whether the settings are correct
            last_hidden_c = self.sources[0].get_last_hidden_state('*')  # Get last c after all blocks
            last_hidden_h = encoder_outputs[input_block_size - 1]  # Get last hidden after the first block

            # Padding so that last hidden_c & _h are the same (this is needed for when using BiLSTM)
            c_shape = tf.shape(last_hidden_c)
            h_shape = tf.shape(last_hidden_h)
            padding = tf.zeros([c_shape[0], h_shape[1] - c_shape[1]])
            last_hidden_c = tf.concat([last_hidden_c, padding], axis=1)

            last_hidden = tf.stack([last_hidden_c, last_hidden_h], axis=0)

        # Note down data
        self.transducer_hidden_units = transducer_hidden_units
        self.num_outputs = n_out
        self.transducer_max_width = transducer_max_width
        self.input_block_size = input_block_size
        self.e_symbol_index = e_symbol_index

        # self.output.placeholder is of shape [transducer_max_width * amount_of_blocks, batch_size, n_out]
        self.output.placeholder = self.build_full_transducer(transducer_hidden_units=transducer_hidden_units,
                                                             embeddings=embeddings,
                                                             num_outputs=n_out,
                                                             input_block_size=input_block_size,
                                                             transducer_max_width=transducer_max_width,
                                                             encoder_outputs=encoder_outputs,
                                                             trans_hidden_init=last_hidden)

        # Set correct logit lengths
        output_size = self.round_vector_to_closest_input_block(vector=self.input_data.size_placeholder[0],
                                                               input_block_size=input_block_size,
                                                               transducer_max_width=transducer_max_width)

        # Set shaping info
        self.output.size_placeholder = {
          0: output_size
        }
        self.output.time_dim_axis = 0
        self.output.batch_dim_axis = 1

        # Add all trainable params
        with self.var_creation_scope() as scope:
            self._add_all_trainable_params(tf.get_collection(tf_compat.v1.GraphKeys.TRAINABLE_VARIABLES, scope=scope.name))

    def build_full_transducer(self, transducer_hidden_units, embeddings, num_outputs, input_block_size,
                              transducer_max_width, encoder_outputs, trans_hidden_init):
        """
        Builds the complete transducer.
        :param int transducer_hidden_units:  Amount of units the transducer should have.
        :param tf.Variable embeddings: Variable with the reference to the embeddings.
        :param int num_outputs: The size of the output layer, i.e. the size of the vocabulary including <E> symbol.
        :param int input_block_size: Amount of inputs to use for each NT block.
        :param int transducer_max_width: The max amount of outputs in one NT block (including the final <E> symbol)
        :param tf.tensor encoder_outputs: The outputs of the encode in shape of [max_time, batch_size, encoder_hidden]
        :param tf.tensor trans_hidden_init: The init state of the transducer. Needs to be of shape
        [2, batch_size, transducer_hidden_units]. The trans_hidden_init[0] is the c vector of the lstm,
        trans_hidden_init[1] the hidden vector.
        :return: Returns a reference to the tf.tensor containing the logits.
        :rtype: tf.tensor
        """

        with self.var_creation_scope():
            # Get meta variables
            batch_size = tf.shape(encoder_outputs)[1]
            if trans_hidden_init is None:
                trans_hidden_init = tf.zeros([2, batch_size, transducer_hidden_units], dtype=tf.float32)

            # Do some more post processing
            max_blocks = tf.cast(tf.shape(encoder_outputs)[0] // input_block_size, tf.int32)
            transducer_list_outputs = tf.ones([max_blocks, batch_size], dtype=tf.int32) * transducer_max_width
            inference_mode = 1.0
            teacher_forcing_targets = tf.ones([transducer_max_width * max_blocks, batch_size], dtype=tf.int32)

            # Process teacher forcing targets
            teacher_forcing_targets_emb = tf.nn.embedding_lookup(embeddings, teacher_forcing_targets)

            # Outputs
            outputs_ta = tf.TensorArray(dtype=tf.float32, size=max_blocks, infer_shape=False)
            init_state = (0, outputs_ta, trans_hidden_init, 0)

            # Init the transducer cell
            from returnn.tf.util.basic import get_initializer
            transducer_cell_initializer = get_initializer('glorot_uniform',
                                                          seed=self.network.random.randint(2 ** 31),
                                                          eval_local_ns={"layer": self})
            transducer_cell = tf.contrib.rnn.LSTMCell(transducer_hidden_units, initializer=transducer_cell_initializer)

            def cond(current_block, outputs_int, trans_hidden, total_output):
                return current_block < max_blocks

            def body(current_block, outputs_int, trans_hidden, total_output):

                # --------------------- TRANSDUCER --------------------------------------------------------------------
                # Each transducer block runs for the max transducer outputs in its respective block

                encoder_raw_outputs = encoder_outputs[input_block_size * current_block:
                                                      input_block_size * (current_block + 1)]

                encoder_raw_outputs = tf.where(tf.is_nan(encoder_raw_outputs), tf.zeros_like(encoder_raw_outputs),
                                               encoder_raw_outputs)

                trans_hidden = tf.where(tf.is_nan(trans_hidden), tf.zeros_like(trans_hidden), trans_hidden)

                # Save/load the state as one tensor, use top encoder layer state as init if this is the first block
                trans_hidden_state = trans_hidden
                transducer_amount_outputs = transducer_list_outputs[current_block]
                transducer_max_output = tf.reduce_max(transducer_amount_outputs)

                # Model building
                helper = tf.contrib.seq2seq.ScheduledEmbeddingTrainingHelper(
                    inputs=teacher_forcing_targets_emb[total_output:total_output + transducer_max_output],  # Get the current target inputs
                    sequence_length=transducer_amount_outputs,
                    embedding=embeddings,
                    sampling_probability=inference_mode,
                    time_major=True
                )

                attention_states = tf.transpose(encoder_raw_outputs,
                                                [1, 0, 2])  # attention_states: [batch_size, max_time, num_enc_units]
                attention_mechanism = tf.contrib.seq2seq.LuongAttention(
                    transducer_hidden_units, attention_states)

                decoder_cell = tf.contrib.seq2seq.AttentionWrapper(
                    transducer_cell,
                    attention_mechanism,
                    attention_layer_size=transducer_hidden_units)

                from tensorflow.python.layers import core as layers_core
                projection_layer = layers_core.Dense(num_outputs, use_bias=False)

                # Build previous state
                trans_hidden_c, trans_hidden_h = tf.split(trans_hidden_state, num_or_size_splits=2, axis=0)
                trans_hidden_c = tf.reshape(trans_hidden_c, shape=[-1, transducer_hidden_units])
                trans_hidden_h = tf.reshape(trans_hidden_h, shape=[-1, transducer_hidden_units])
                from tensorflow.contrib.rnn import LSTMStateTuple
                trans_hidden_state_t = LSTMStateTuple(trans_hidden_c, trans_hidden_h)

                decoder = tf.contrib.seq2seq.BasicDecoder(
                    decoder_cell, helper,
                    decoder_cell.zero_state(batch_size, tf.float32).clone(cell_state=trans_hidden_state_t),
                    output_layer=projection_layer)
                outputs, transducer_hidden_state_new, _ = tf.contrib.seq2seq.dynamic_decode(decoder,
                                                                                            output_time_major=True,
                                                                                            maximum_iterations=transducer_max_output)
                logits = outputs.rnn_output  # logits of shape [max_time,batch_size,vocab_size]

                # Modify output of transducer_hidden_state_new so that it can be fed back in again without problems.

                transducer_hidden_state_new = tf.concat(
                    [transducer_hidden_state_new[0].c, transducer_hidden_state_new[0].h],
                    axis=0)
                transducer_hidden_state_new = tf.reshape(transducer_hidden_state_new,
                                                         shape=[2, -1, transducer_hidden_units])

                # Note the outputs
                outputs_int = outputs_int.write(current_block, logits)

                return current_block + 1, outputs_int, \
                    transducer_hidden_state_new, total_output + transducer_max_output

            _, outputs_final, _, _ = tf.while_loop(cond, body, init_state, parallel_iterations=1)

            # Process outputs
            with tf.device('/cpu:0'):
              logits = outputs_final.concat()  # And now it's [max_output_time, batch_size, num_outputs]

            # For loading the model later on
            logits = tf.identity(logits, name='logits')

        return logits

    def _add_all_trainable_params(self, tf_vars):
        for var in tf_vars:
            self.add_param(param=var, trainable=True, saveable=True)

    def round_vector_to_closest_input_block(self, vector, input_block_size, transducer_max_width):
        """
        Rounds up the provided vector so that every entry is a multiple of input_block_size.
        :param tf.tensor vector: A vector.
        :param int input_block_size: Input block size as specified in the __init__ function.
        :return: tf.tensor A vector the same shape as 'vector'.
        """
        vector = tf.cast(tf_compat.v1.ceil(tf.cast(vector, tf.float32) / input_block_size), tf.float32) * tf.cast(transducer_max_width, tf.float32)
        vector = tf.cast(vector, tf.int32)

        return vector


    @classmethod
    def get_out_data_from_opts(cls, n_out, **kwargs):
        data = get_concat_sources_data_template(kwargs["sources"], name="%s_output" % kwargs["name"])
        data = data.copy_as_time_major()  # type: Data
        data.shape = (None, n_out)
        data.time_dim_axis = 0
        data.batch_dim_axis = 1
        data.dim = n_out
        return data


class NeuralTransducerLoss(Loss):
    """
    The loss function that should be used with the NeuralTransducer layer. This loss function has the built in
    alignment algorithm from the original paper.
    """
    class_name = "neural_transducer"
    recurrent = True

    class Alignment(object):
        """
        Class to manage the alignment generation in the NT.
        """

        def __init__(self, transducer_hidden_units, E_SYMBOL):
            """
            Alignment initiation.
            :param int transducer_hidden_units: Amount of hidden units that the transducer should have.
            :param int E_SYMBOL: The index of the <e> symbol.
            """
            import numpy
            self.alignment_position = (0, 1)  # first entry is position in target (y~), second is the block index
            self.log_prob = 0  # The sum log prob of this alignment over the target indices
            self.alignment_locations = []  # At which indices in the target output we need to insert <e>
            self.last_state_transducer = numpy.zeros(
                shape=(2, 1, transducer_hidden_units))  # Transducer state
            self.E_SYMBOL = E_SYMBOL  # Index of

        def __compute_sum_probabilities(self, transducer_outputs, targets, transducer_amount_outputs):
            """
            # TODO move this function outside the Alignment class scope
            Computes the sum log probabilities of the outputs based on the targets.
            :param np.ndarray[int] transducer_outputs: Softmaxed transducer outputs of one block.
            Size: [transducer_amount_outputs, 1, num_outputs]
            :param [int] targets: List of targets.
            :param int transducer_amount_outputs: The width of this transducer block.
            :return: The summed log prob for this block.
            :rtype: float
            """
            import numpy

            def get_prob_at_timestep(timestep):
                if timestep + start_index < len(targets):
                    # For normal operations
                    if transducer_outputs[timestep][0][targets[start_index + timestep]] <= 0:
                        return -10000000.0 + numpy.random.uniform(-100, -500)  # Some large negative number
                    else:
                        return numpy.log(transducer_outputs[timestep][0][targets[start_index + timestep]])
                else:
                    # For last timestep, so the <e> symbol
                    if transducer_outputs[timestep][0][self.E_SYMBOL] <= 0:
                        return -10000000.0 + numpy.random.uniform(-100, -500)  # Some large negative number
                    else:
                        return numpy.log(transducer_outputs[timestep][0][self.E_SYMBOL])

            # print transducer_outputs
            start_index = self.alignment_position[
                              0] - transducer_amount_outputs  # The current position of this alignment
            prob = 0
            for i in range(0,
                           transducer_amount_outputs + 1):  # Do not include e symbol in calculation, +1 due to last symbol
                prob += get_prob_at_timestep(i)
            return prob

        def insert_alignment(self, index, block_index, transducer_outputs, targets, transducer_amount_outputs,
                             new_transducer_state):
            """
            Inserts alignment properties for a new block.
            :param int index: The index of of y~ corresponding to the last target index.
            :param int block_index: The new block index.
            :param np.ndarray transducer_outputs: The computed transducer outputs. Shape
            [transducer_amount_outputs, 1, n_out]
            :param np.ndarray targets: The complete target array, should be of shape [total_target_length].
            :param int transducer_amount_outputs: The amount of outputs that the transducer created in this block.
            :param np.ndarray new_transducer_state: The new transducer state of shape [2, 1, transducer_hidden_units]
            """
            self.alignment_locations.append(index)
            self.alignment_position = (index, block_index)
            self.log_prob += self.__compute_sum_probabilities(transducer_outputs, targets, transducer_amount_outputs)
            self.last_state_transducer = new_transducer_state

    def __init__(self, debug=False, max_variance=999999.9, **kwargs):
        """
        Initialize the Neural Transducer loss.
        :param bool debug: Whether to output debug info such as alignments, argmax, variance etc...
        :param float max_variance: If a time step (in CE) has a too high variance in within the batch, then the gradient
        for that time step will be ignored. Set this value lower if you have outliers that disrupt training.
        """
        super(NeuralTransducerLoss, self).__init__(**kwargs)

        self.transducer_hidden_units = 0
        self.num_outputs = 0
        self.transducer_max_width = 0
        self.input_block_size = 0
        self.e_symbol_index = 0
        self.debug = debug
        self.reduce_func = tf.reduce_sum
        self.max_variance = max_variance

    def init(self, **kwargs):
        super(NeuralTransducerLoss, self).init(**kwargs)
        # Get setup vars from sources
        base_class = None
        for c in self.base_network.layers:
            if isinstance(self.base_network.layers[c], NeuralTransducerLayer):
                base_class = self.base_network.layers[c]

        assert base_class is not None, "Neural Transducer layer not found!"
        self.transducer_hidden_units = base_class.transducer_hidden_units
        self.num_outputs = base_class.num_outputs
        self.transducer_max_width = base_class.transducer_max_width
        self.input_block_size = base_class.input_block_size
        self.e_symbol_index = base_class.e_symbol_index

    def get_value(self):
        logits = self.output.copy_as_time_major().placeholder
        logits_lengths = self.output.size_placeholder[0]
        targets = self.target.copy_as_time_major().placeholder
        targets_lengths = self.target.size_placeholder[0]

        # Get alignment info into our targets
        new_targets, mask = tf_compat.v1.py_func(func=self.get_alignment_from_logits_manager,
                                                 inp=[logits, targets, logits_lengths, targets_lengths],
                                                 Tout=(tf.int64, tf.bool), stateful=False)

        # Get CE
        stepwise_cross_entropy = tf.nn.sparse_softmax_cross_entropy_with_logits(labels=new_targets, logits=logits)

        # Debugging
        if self.debug is True:
            stepwise_cross_entropy = tf.Print(stepwise_cross_entropy, [targets[:, 0]], message='Targets: ',
                                              summarize=1000)
            stepwise_cross_entropy = tf.Print(stepwise_cross_entropy, [new_targets[:, 0]],
                                              message='Targets with alignment: ', summarize=1000)
            stepwise_cross_entropy = tf.Print(stepwise_cross_entropy, [tf.argmax(logits, axis=2)[:, 0]],
                                              message='Argmax: ', summarize=1000)

        # Apply masking step AFTER cross entropy:
        zeros = tf.zeros_like(stepwise_cross_entropy)
        stepwise_cross_entropy = tf.where(mask, stepwise_cross_entropy, zeros)

        # Check for outliers and set their gradient to 0
        loss_time = tf.reduce_sum(stepwise_cross_entropy, axis=1)
        mean, variance = tf_compat.v1.nn.moments(stepwise_cross_entropy, axes=[1])
        loss_mask = tf.cast(tf.greater(variance, self.max_variance), tf.float32)
        stepwise_cross_entropy = (
          tf.stop_gradient(tf.multiply(loss_mask, loss_time)) +
          tf.multiply(tf.cast(tf.logical_not(tf.cast(loss_mask, tf.bool)), tf.float32), loss_time))

        if self.debug is True:
            stepwise_cross_entropy = tf.cond(tf.reduce_sum(loss_mask) >= 1,
                                             lambda: tf.Print(stepwise_cross_entropy, [variance, loss_mask],
                                                              message='High Variance: ', summarize=500),
                                             lambda: stepwise_cross_entropy)

        # TODO: add forwarding layer

        # Get full loss
        norm = tf.to_float(tf.reduce_sum(targets_lengths)) / tf.reduce_sum(tf.to_float(mask))
        loss = tf.reduce_sum(stepwise_cross_entropy) * norm

        return loss

    def get_alignment_from_logits(self, logits, targets, amount_of_blocks, transducer_max_width):
        """
        Finds the alignment of the target sequence to the actual output.
        :param logits: Logits from transducer, of size [transducer_max_width * amount_of_blocks, 1, vocab_size]
        :param targets: The target sequence of shape [time] where each entry is an index.
        :param amount_of_blocks: Amount of blocks in Neural Transducer.
        :param transducer_max_width: The max width of one transducer block.
        :return: Returns a list of indices where <e>'s need to be inserted into the target sequence, shape: [max_time, 1]
        (see paper) and a boolean mask for use with a loss function of shape [max_time, 1].
        """
        import numpy
        import copy
        # Split logits into list of arrays with each array being one block
        # of shape [transducer_max_width, 1, vocab_size]
        logits = numpy.reshape(logits, newshape=[logits.shape[0], 1, logits.shape[1]])

        split_logits = numpy.split(logits, amount_of_blocks)

        # print 'Raw logits: ' + str(softmax(split_logits[0][0:transducer_max_width], axis=2))

        def run_new_block(previous_alignments, block_index, transducer_max_width, targets,
                          total_blocks):
            """
            Runs one block of the alignment process.
            :param previous_alignments: List of alignment objects from previous block step.
            :param block_index: The index of the current new block.
            :param transducer_max_width: The max width of the transducer block.
            :param targets: The full target array of shape [time]
            :param total_blocks: The total amount of blocks.
            :return: new_alignments as list of Alignment objects
            """

            def run_transducer(current_block, transducer_width):
                # apply softmax on the correct outputs
                transducer_out = softmax(split_logits[current_block][0:transducer_width], axis=2)
                return transducer_out

            # Look into every existing alignment
            new_alignments = []
            for i in range(len(previous_alignments)):
                alignment = previous_alignments[i]

                # Expand the alignment for each transducer width, only look at valid options
                targets_length = len(targets)

                min_index = alignment.alignment_position[0] + transducer_max_width + \
                            max(-transducer_max_width,
                                targets_length - ((total_blocks - block_index + 1) * transducer_max_width
                                                  + alignment.alignment_position[0]))
                max_index = alignment.alignment_position[0] + transducer_max_width + min(0, targets_length - (
                    alignment.alignment_position[0] + transducer_max_width))

                # new_alignment_index's value is equal to the index of y~ for that computation
                for new_alignment_index in range(min_index, max_index + 1):  # 1 so that the max_index is also used
                    # Create new alignment
                    new_alignment = copy.deepcopy(alignment)
                    new_alignment_width = new_alignment_index - new_alignment.alignment_position[0]
                    trans_out = run_transducer(transducer_width=new_alignment_width + 1, current_block=block_index - 1)

                    new_alignment.insert_alignment(new_alignment_index, block_index, trans_out, targets,
                                                   new_alignment_width, None)
                    new_alignments.append(new_alignment)

            # Delete all overlapping alignments, keeping the highest log prob
            for a in reversed(new_alignments):
                for o in new_alignments:
                    if o is not a and a.alignment_position == o.alignment_position and o.log_prob > a.log_prob:
                        if a in new_alignments:
                            new_alignments.remove(a)

            assert len(new_alignments) > 0, 'Error in amount of alignments! %s' % str(targets)

            return new_alignments

        # Manage variables
        current_block_index = 1
        current_alignments = [self.Alignment(transducer_hidden_units=self.transducer_hidden_units,
                                             E_SYMBOL=self.e_symbol_index)]

        # Do assertions to check whether everything was correctly set up.
        assert (transducer_max_width - 1) * amount_of_blocks >= len(
            targets), 'transducer_max_width to small for targets'

        for block in range(current_block_index, amount_of_blocks + 1):
            # Run all blocks
            current_alignments = run_new_block(previous_alignments=current_alignments,
                                               block_index=block,
                                               transducer_max_width=transducer_max_width - 1,  # -1 due to offset for e
                                               targets=targets, total_blocks=amount_of_blocks)
            # for alignment in current_alignments:

        # Select first alignment if we have multiple with the same log prob (happens with ~1% probability in training)
        if self.debug is True:
            print('Alignment: ' + str(current_alignments[0].alignment_locations) + ' for targets: ' + str(targets))

        def modify_targets(targets, alignment):
            # Calc lengths for each transducer block
            lengths_temp = []
            alignment.insert(0, 0)  # This is so that the length calculation is done correctly
            for i in range(1, len(alignment)):
                lengths_temp.append(alignment[i] - alignment[i - 1] + 1)
            del alignment[0]  # Remove alignment index that we added
            lengths = lengths_temp

            # Modify targets so that it has the appropriate alignment
            offset = 0
            for e in alignment:
                targets.insert(e + offset, self.e_symbol_index)
                offset += 1

            # Modify so that all targets have same lengths in each transducer block using 0 (will be masked away)
            offset = 0
            for i in range(len(alignment)):
                for app in range(transducer_max_width - lengths[i]):
                    targets.insert(offset + lengths[i], 0)
                offset += transducer_max_width

            # Process targets back to time major
            targets = numpy.asarray([targets])
            targets = numpy.transpose(targets, axes=[1, 0])

            return targets, lengths

        m_targets, lengths = modify_targets(targets.tolist(), current_alignments[0].alignment_locations)
        # m_targets now of shape: [max_time, 1 (batch_size)] = [transducer_max_width * number_of_blocks, 1]

        # Create boolean mask for TF so that unnecessary logits are not used for the loss function
        # Of shape [max_time, batch_size], True where gradient data is kept, False where not

        def create_mask(lengths):
            mask = numpy.full(m_targets.shape, False)
            for i in range(amount_of_blocks):
                for j in range(lengths[i]):
                    mask[i*transducer_max_width:i*transducer_max_width + j + 1, 0] = True
            return mask

        mask = create_mask(lengths)

        return m_targets, mask

    def get_alignment_from_logits_manager(self, logits, targets, logit_lengths, targets_lengths):
        """
        Get the modified targets & mask.
        :param logits: Logits of shape [max_time, batch_size, vocab_size]
        :param targets: Targets of shape [max_time, batch_size]. Each entry denotes the index of the correct target.
        :return: modified targets of shape [max_time, batch_size, vocab_size]
        & mask of shape [max_time, batch_size]
        """
        import numpy
        logits = numpy.copy(logits)
        targets = numpy.copy(targets)

        # print('Targets: ' + str(targets), file=log.v1)

        m_targets = []
        masks = []

        # amount_of_blocks = int(logits.shape[0]/self.transducer_max_width)

        # Go over every sequence in batch
        for batch_index in range(logits.shape[1]):
            # Slice correct logits & targets
            logit_length = logit_lengths[batch_index]
            target_length = targets_lengths[batch_index]
            amount_of_blocks = int(logit_length/self.transducer_max_width)

            temp_target, temp_mask = self.get_alignment_from_logits(logits=logits[0:logit_length, batch_index, :],
                                                                    targets=targets[0:target_length, batch_index],
                                                                    amount_of_blocks=amount_of_blocks,
                                                                    transducer_max_width=self.transducer_max_width)
            # Pad afterwards each target (based on targets_lengths) & mask (based on logit_lengths)
            temp_target = numpy.append(temp_target, numpy.zeros(shape=(logits.shape[0] - temp_target.shape[0], 1), dtype=int), axis=0)
            temp_mask = numpy.append(temp_mask, numpy.zeros(shape=(logits.shape[0] - logit_length, 1), dtype=bool), axis=0)

            m_targets.append(temp_target)
            masks.append(temp_mask)

        # Concatenate the targets & masks on the time axis; due to padding m_targets are all the same
        m_targets = numpy.concatenate(m_targets, axis=1)
        masks = numpy.concatenate(masks, axis=1)

        return m_targets, masks

    @classmethod
    def get_auto_output_layer_dim(cls, target_dim):
        return target_dim + 1  # one added for <E>

    def get_error(self):
        with tf.name_scope("loss_frame_error"):
            logits = self.output.copy_as_time_major().placeholder
            logits_lengths = self.output.size_placeholder[0]
            targets = self.target.copy_as_time_major().placeholder
            targets_lengths = self.target.size_placeholder[0]

            # Get alignment info into our targets
            new_targets, mask = tf_compat.v1.py_func(func=self.get_alignment_from_logits_manager,
                                                     inp=[logits, targets, logits_lengths, targets_lengths],
                                                     Tout=(tf.int64, tf.bool), stateful=False)

            output_label = tf.cast(tf.argmax(logits, axis=2), tf.int64)
            zeros = tf.zeros_like(output_label)

            # Calculate edit distance
            # First modify outputs so that only those outputs in the mask are considered
            mod_logits = tf.where(mask, output_label, zeros)

            # Get find seq lens (due to having blank spaces in the modified targets we need to use this method to get
            # the correct seq lens)
            seq_lens = tf.argmax(tf.cumsum(tf.cast(mask, tf.int32), axis=0), axis=0)
            seq_lens = tf.reshape(seq_lens, shape=[tf.shape(seq_lens)[0]])

            logits_sparse = sparse_labels_with_seq_lens(tf.transpose(mod_logits), seq_lens=seq_lens)
            targets_sparse = sparse_labels_with_seq_lens(tf.transpose(new_targets), seq_lens=seq_lens)

            e = tf.edit_distance(logits_sparse[0], targets_sparse[0], normalize=False)
            total = tf.reduce_sum(e)

            norm = tf.cast(tf.reduce_sum(targets_lengths), tf.float32) / tf.reduce_sum(tf.cast(mask, tf.float32))
            total = total * norm

            return total



#!/usr/bin/env python3

"""
Goes through a Sprint Bliss XML file,
and writes out sequence lengths based on the start/end time,
into a Python-formatted file.
"""

from __future__ import print_function

import sys
from argparse import ArgumentParser
import gzip
from xml.etree import ElementTree
import itertools
import _setup_returnn_env  # noqa


class BlissItem:
  """
  Represents one entry in the Bliss XML.
  """
  def __init__(self, segment_name, recording_filename, start_time, end_time, orth):
    """
    :param str segment_name:
    :param str recording_filename:
    :param float start_time:
    :param float end_time:
    :param str orth:
    """
    self.segment_name = segment_name
    self.recording_filename = recording_filename
    self.start_time = start_time
    self.end_time = end_time
    self.orth = orth

  def __repr__(self):
    keys = ["segment_name", "recording_filename", "start_time", "end_time", "orth"]
    return "BlissItem(%s)" % ", ".join(["%s=%r" % (key, getattr(self, key)) for key in keys])

  @property
  def delta_time(self):
    """
    :rtype: float
    """
    return self.end_time - self.start_time


def iter_bliss(filename):
  """
  :param str filename:
  :return: yields BlissItem
  :rtype: list[BlissItem]
  """
  corpus_file = open(filename, 'rb')
  if filename.endswith(".gz"):
    corpus_file = gzip.GzipFile(fileobj=corpus_file)

  context = iter(ElementTree.iterparse(corpus_file, events=('start', 'end')))
  _, root = next(context)  # get root element
  name_tree = [root.attrib["name"]]
  elem_tree = [root]
  count_tree = [0]
  recording_filename = None
  for event, elem in context:
    if elem.tag == "recording":
      recording_filename = elem.attrib["audio"] if event == "start" else None
    if event == 'end' and elem.tag == "segment":
      elem_orth = elem.find("orth")
      orth_raw = elem_orth.text  # should be unicode
      orth_split = orth_raw.split()
      orth = " ".join(orth_split)
      segment_name = "/".join(name_tree)
      yield BlissItem(
        segment_name=segment_name, recording_filename=recording_filename,
        start_time=float(elem.attrib["start"]), end_time=float(elem.attrib["end"]),
        orth=orth)
      root.clear()  # free memory
    if event == "start":
      count_tree[-1] += 1
      count_tree.append(0)
      elem_tree += [elem]
      elem_name = elem.attrib.get("name", None)
      if elem_name is None:
        elem_name = str(count_tree[-2])
      assert isinstance(elem_name, str)
      name_tree += [elem_name]
    elif event == "end":
      assert elem_tree[-1] is elem
      elem_tree = elem_tree[:-1]
      name_tree = name_tree[:-1]
      count_tree = count_tree[:-1]


def main():
  """
  Main entry.
  """
  arg_parser = ArgumentParser()
  arg_parser.add_argument("bliss_filename", nargs="+")
  arg_parser.add_argument("--output", default="/dev/stdout")
  args = arg_parser.parse_args()
  if args.output.endswith(".gz"):
    out = gzip.GzipFile(args.output, mode="wb")
  else:
    out = open(args.output, "wb")
  out.write(b"{\n")
  for bliss_item in itertools.chain(*[iter_bliss(fn) for fn in args.bliss_filename]):
    assert isinstance(bliss_item, BlissItem)
    seq_len = round(bliss_item.delta_time * 100.)  # assume 10ms frames, round
    out.write(b"%r: %i,\n" % (bliss_item.segment_name, seq_len))
  out.write(b"}\n")
  out.close()


if __name__ == "__main__":
  from returnn.util import better_exchook
  better_exchook.install()
  try:
    main()
  except BrokenPipeError:
    print("BrokenPipeError")
    sys.exit(1)



#!/usr/bin/env python

"""
Collect orth symbols.
"""

from __future__ import print_function

import os
import sys

import _setup_returnn_env  # noqa
import returnn.__main__ as rnn
from returnn.log import log
from returnn.config import Config
import argparse
from returnn.util.basic import hms, human_size, parse_orthography, parse_orthography_into_symbols, unicode
import gzip
from xml.etree import ElementTree
import wave
import time


def found_sub_seq(sub_seq, seq):
  """
  :param list[str] sub_seq:
  :param list[str] seq:
  :rtype: bool
  """
  # Very inefficient naive implementation:
  for i in range(len(seq)):
    if seq[i:i+len(sub_seq)] == sub_seq:
      return True
  return False


def iter_dataset(dataset, options, callback):
  """
  :type dataset: Dataset.Dataset
  """
  dataset.init_seq_order(epoch=1)
  assert "orth" in dataset.get_target_list()

  seq_idx = 0
  while dataset.is_less_than_num_seqs(seq_idx):
    dataset.load_seqs(seq_idx, seq_idx)

    frame_len = dataset.get_seq_length(seq_idx)["data"]
    orth = dataset.get_targets("orth", seq_idx)
    callback(frame_len=frame_len, orth=orth)

    seq_idx += 1


def get_wav_time_len(filename):
  """
  :param str filename:
  :rtype: float
  """
  f = wave.open(filename)
  num_frames = f.getnframes()
  frame_rate = f.getframerate()
  f.close()
  return num_frames / float(frame_rate)


def iter_bliss(filename, options, callback):
  corpus_file = open(filename, 'rb')
  if filename.endswith(".gz"):
    corpus_file = gzip.GzipFile(fileobj=corpus_file)

  def getelements(tag):
    """Yield *tag* elements from *filename_or_file* xml incrementaly."""
    context = iter(ElementTree.iterparse(corpus_file, events=('start', 'end')))
    _, root = next(context) # get root element
    tree = [root]
    for event, elem in context:
      if event == "start":
        tree += [elem]
      elif event == "end":
        assert tree[-1] is elem
        tree = tree[:-1]
      if event == 'end' and elem.tag == tag:
        yield tree, elem
        root.clear() # free memory

  time_via_wav = False

  for tree, elem in getelements("segment"):
    if options.collect_time:
      start = float(elem.attrib.get('start', 0))
      if "end" in elem.attrib:
        end = float(elem.attrib['end'])
      else:
        if not time_via_wav:
          time_via_wav = True
          print("Time will be read from WAV recordings. Can be slow. Maybe use `--collect_time False`.", file=log.v3)
        rec_elem = tree[-1]
        assert rec_elem.tag == "recording"
        wav_filename = rec_elem.attrib["audio"]
        end = get_wav_time_len(wav_filename)
      assert end > start
      frame_len = (end - start) * (1000.0 / options.frame_time)
    else:
      frame_len = 0
    elem_orth = elem.find("orth")
    orth_raw = elem_orth.text  # should be unicode
    orth_split = orth_raw.split()
    orth = " ".join(orth_split)

    callback(frame_len=frame_len, orth=orth)


def iter_txt(filename, options, callback):
  f = open(filename, 'rb')
  if filename.endswith(".gz"):
    f = gzip.GzipFile(fileobj=f)

  if options.collect_time:
    print("No time-info in txt.", file=log.v3)
    options.collect_time = False

  for line in f:
    line = line.strip()
    if not line:
      continue

    callback(frame_len=0, orth=line)


def collect_stats(options, iter_corpus):
  """
  :param options: argparse.Namespace
  """
  orth_symbols_filename = options.output
  if orth_symbols_filename:
    assert not os.path.exists(orth_symbols_filename)

  class Stats:
    count = 0
    process_last_time = time.time()
    total_frame_len = 0
    total_orth_len = 0
    orth_syms_set = set()

  if options.add_numbers:
    Stats.orth_syms_set.update(map(chr, list(range(ord("0"), ord("9") + 1))))
  if options.add_lower_alphabet:
    Stats.orth_syms_set.update(map(chr, list(range(ord("a"), ord("z") + 1))))
  if options.add_upper_alphabet:
    Stats.orth_syms_set.update(map(chr, list(range(ord("A"), ord("Z") + 1))))

  def cb(frame_len, orth):
    if frame_len >= options.max_seq_frame_len:
      return
    orth_syms = parse_orthography(orth)
    if len(orth_syms) >= options.max_seq_orth_len:
      return

    Stats.count += 1
    Stats.total_frame_len += frame_len

    if options.dump_orth_syms:
      print("Orth:", "".join(orth_syms), file=log.v3)
    if options.filter_orth_sym:
      if options.filter_orth_sym in orth_syms:
        print("Found orth:", "".join(orth_syms), file=log.v3)
    if options.filter_orth_syms_seq:
      filter_seq = parse_orthography_into_symbols(options.filter_orth_syms_seq)
      if found_sub_seq(filter_seq, orth_syms):
        print("Found orth:", "".join(orth_syms), file=log.v3)
    Stats.orth_syms_set.update(orth_syms)
    Stats.total_orth_len += len(orth_syms)

    # Show some progress if it takes long.
    if time.time() - Stats.process_last_time > 2:
      Stats.process_last_time = time.time()
      if options.collect_time:
        print(
          "Collect process, total frame len so far:", hms(Stats.total_frame_len * (options.frame_time / 1000.0)),
          file=log.v3)
      else:
        print("Collect process, total orth len so far:", human_size(Stats.total_orth_len), file=log.v3)

  iter_corpus(cb)

  if options.remove_symbols:
    filter_syms = parse_orthography_into_symbols(options.remove_symbols)
    Stats.orth_syms_set -= set(filter_syms)

  if options.collect_time:
    print(
      "Total frame len:", Stats.total_frame_len, "time:", hms(Stats.total_frame_len * (options.frame_time / 1000.0)),
      file=log.v3)
  else:
    print("No time stats (--collect_time False).", file=log.v3)
  print("Total orth len:", Stats.total_orth_len, "(%s)" % human_size(Stats.total_orth_len), end=' ', file=log.v3)
  if options.collect_time:
    print("fraction:", float(Stats.total_orth_len) / Stats.total_frame_len, file=log.v3)
  else:
    print("", file=log.v3)
  print("Average orth len:", float(Stats.total_orth_len) / Stats.count, file=log.v3)
  print("Num symbols:", len(Stats.orth_syms_set), file=log.v3)

  if orth_symbols_filename:
    orth_syms_file = open(orth_symbols_filename, "wb")
    for orth_sym in sorted(Stats.orth_syms_set):
      orth_syms_file.write(b"%s\n" % unicode(orth_sym).encode("utf8"))
    orth_syms_file.close()
    print("Wrote orthography symbols to", orth_symbols_filename, file=log.v3)
  else:
    print("Provide --output to save the symbols.", file=log.v3)


def init(config_filename=None):
  rnn.init_better_exchook()
  rnn.init_thread_join_hack()
  if config_filename:
    rnn.init_config(config_filename, command_line_options=[])
    rnn.init_log()
  else:
    log.initialize()
  print("RETURNN collect-orth-symbols starting up.", file=log.v3)
  rnn.init_faulthandler()
  if config_filename:
    rnn.init_config_json_network()
    rnn.init_data()
    rnn.print_task_properties()


def is_bliss(filename):
  try:
    corpus_file = open(filename, 'rb')
    if filename.endswith(".gz"):
      corpus_file = gzip.GzipFile(fileobj=corpus_file)
    context = iter(ElementTree.iterparse(corpus_file, events=('start', 'end')))
    _, root = next(context)  # get root element
    return True
  except IOError:  # 'Not a gzipped file' or so
    pass
  except ElementTree.ParseError:  # 'syntax error' or so
    pass
  return False


def is_crnn_config(filename):
  if filename.endswith(".gz"):
    return False
  try:
    config = Config()
    config.load_file(filename)
    return True
  except Exception:
    pass
  return False


def main(argv):
  argparser = argparse.ArgumentParser(description='Collect orth symbols.')
  argparser.add_argument('input', help="RETURNN config, Corpus Bliss XML or just txt-data")
  argparser.add_argument(
    '--frame_time', type=int, default=10, help='time (in ms) per frame. not needed for Corpus Bliss XML')
  argparser.add_argument('--collect_time', type=int, default=True, help="collect time info. can be slow in some cases")
  argparser.add_argument('--dump_orth_syms', action='store_true', help="dump all orthographies")
  argparser.add_argument('--filter_orth_sym', help="dump orthographies which match this filter")
  argparser.add_argument('--filter_orth_syms_seq', help="dump orthographies which match this filter")
  argparser.add_argument(
    '--max_seq_frame_len', type=int, default=float('inf'), help="collect only orthographies <= this max frame len")
  argparser.add_argument(
    '--max_seq_orth_len', type=int, default=float('inf'), help="collect only orthographies <= this max orth len")
  argparser.add_argument('--add_numbers', type=int, default=True, help="add chars 0-9 to orth symbols")
  argparser.add_argument('--add_lower_alphabet', type=int, default=True, help="add chars a-z to orth symbols")
  argparser.add_argument('--add_upper_alphabet', type=int, default=True, help="add chars A-Z to orth symbols")
  argparser.add_argument('--remove_symbols', default="(){}$", help="remove these chars from orth symbols")
  argparser.add_argument('--output', help='where to store the symbols (default: dont store)')
  args = argparser.parse_args(argv[1:])

  bliss_filename = None
  crnn_config_filename = None
  txt_filename = None
  if is_bliss(args.input):
    bliss_filename = args.input
  elif is_crnn_config(args.input):
    crnn_config_filename = args.input
  else:  # treat just as txt
    txt_filename = args.input
  init(config_filename=crnn_config_filename)

  if bliss_filename:
    iter_corpus = lambda cb: iter_bliss(bliss_filename, options=args, callback=cb)
  elif txt_filename:
    iter_corpus = lambda cb: iter_txt(txt_filename, options=args, callback=cb)
  else:
    iter_corpus = lambda cb: iter_dataset(rnn.train_data, options=args, callback=cb)
  collect_stats(args, iter_corpus)

  if crnn_config_filename:
    rnn.finalize()


if __name__ == '__main__':
  main(sys.argv)




"""
This file is going to be imported by Debug.debug_shell() and available as interactive commands.
"""

import sys
import numpy
import h5py
from returnn.util import BackendEngine

if BackendEngine.is_theano_selected():
  # noinspection PyUnresolvedReferences,PyPackageRequirements
  import theano
  # noinspection PyUnresolvedReferences,PyPackageRequirements
  import theano.tensor as tt
  # noinspection PyUnresolvedReferences,PyPackageRequirements
  import theano.sandbox.cuda as cuda
  from returnn.theano.util import make_var_tuple
  from returnn.theano.network import LayerNetwork
else:
  theano = None


def find_obj_in_stack(cls, stack=None, all_threads=True):
  """
  :param type cls:
  :param types.FrameType|traceback.FrameSummary stack:
  :param bool all_threads:
  :return: obj
  """
  if all_threads:
    assert stack is None
    # noinspection PyProtectedMember,PyUnresolvedReferences
    for tid, stack in sys._current_frames().items():
      obj = find_obj_in_stack(cls=cls, stack=stack, all_threads=False)
      if obj is not None:
        return obj
    return None

  assert not all_threads
  if stack is None:
    # noinspection PyProtectedMember,PyUnresolvedReferences
    stack = sys._getframe()
    assert stack, "could not get stack"

  import inspect
  isframe = inspect.isframe

  _tb = stack
  while _tb is not None:
    if isframe(_tb):
      f = _tb
    else:
      f = _tb.tb_frame

    for obj in f.f_locals.values():
      if isinstance(obj, cls):
        return obj

    if isframe(_tb):
      _tb = _tb.f_back
    else:
      _tb = _tb.tb_next

  return None


_device = None


def get_device():
  """
  :rtype: Device.Device
  """
  global _device
  if _device:
    return _device
  from returnn.theano.device import Device
  _device = find_obj_in_stack(Device)
  return _device


if BackendEngine.is_theano_selected():
  def compute(var, trainnet=True):
    """
    :param theano.Variable var: variable which we should compute the value of
    :param bool trainnet: whether to make givens based on dev.trainnet or dev.testnet
    :return: the computed value
    :rtype: numpy.ndarray

    This expects to calculate some value of the trainnet or testnet of the current Device.
    """
    dev = get_device()
    assert dev, "no Device instance found"
    if trainnet:
      network = dev.trainnet
    else:
      network = dev.testnet
    givens = dev.make_givens(network)
    if isinstance(var, list):
      outputs = var
    else:
      outputs = [var]
    func = theano.function(inputs=[dev.block_start, dev.block_end],
                           outputs=outputs,
                           givens=givens,
                           on_unused_input='warn',
                           name="debug compute")
    batch_dim = dev.y["data"].get_value(borrow=True, return_internal_type=True).shape[1]
    batch_start = 0
    batch_end = batch_dim
    result = func(batch_start, batch_end)
    if not isinstance(var, list):
      result = result[0]
    return result

  class DebugNn:
    """
    Debug helper for NN.
    """

    def __init__(self, filename):
      self.network = LayerNetwork.from_hdf(filename, mask="unity", train_flag=False, eval_flag=True)
      self.f_forwarder = None

    def compile_forwarder(self):
      """
      Compile function for forwarding.
      """
      network = self.network
      data_keys = list(sorted(network.j.keys()))
      # All input seqs expected to have same length.
      givens = [(network.j[k], tt.ones(network.y["data"].shape[:2], dtype="int8")) for k in data_keys]
      self.f_forwarder = theano.function(
        inputs=[network.y["data"]],
        outputs=(
          [network.output["output"].output] +
          [layer.output for name, layer in sorted(network.output.items()) if name != "output"]),
        givens=givens,
        on_unused_input='warn',
        name="forwarder")

    def forward(self, data, output_index=0):
      """
      :param numpy.ndarray data: shape (time,in-dim)
      :param int output_index:
      :return: shape (time,out-dim)
      :rtype: numpy.ndarray
      """
      assert data.ndim == 2
      data = data[:, None, :]  # add batch-dim
      assert self.f_forwarder
      res = self.f_forwarder(data)
      res = make_var_tuple(res)[output_index]
      assert res.ndim == 3
      assert res.shape[1] == 1
      res = res[:, 0]
      return res


class SimpleHdf:
  """
  Simple HDF writer.
  """

  def __init__(self, filename):
    self.hdf = h5py.File(filename)
    self.seq_tag_to_idx = {name: i for (i, name) in enumerate(self.hdf["seqTags"])}
    self.num_seqs = len(self.hdf["seqTags"])
    assert self.num_seqs == len(self.seq_tag_to_idx), "not unique seq tags"
    seq_lens = self.hdf["seqLengths"]
    if seq_lens.ndim == 2:
      seq_lens = seq_lens[:, 0]
    assert self.num_seqs == len(seq_lens)
    self.seq_starts = [0] + list(numpy.cumsum(seq_lens))
    total_len = self.seq_starts[-1]
    inputs_len = self.hdf["inputs"].shape[0]
    assert total_len == inputs_len, "time-dim does not match: %i vs %i" % (total_len, inputs_len)
    assert self.seq_starts[-1] == self.hdf["targets/data/classes"].shape[0]

  def get_seq_tags(self):
    """
    :rtype: list[str]
    """
    return self.hdf["seqTags"]

  def get_data(self, seq_idx):
    """
    :param int seq_idx:
    :rtype: numpy.ndarray
    """
    seq_t0, seq_t1 = self.seq_starts[seq_idx:seq_idx + 2]
    return self.hdf["inputs"][seq_t0:seq_t1]

  def get_targets(self, seq_idx):
    """
    :param int seq_idx:
    :rtype: numpy.ndarray
    """
    seq_t0, seq_t1 = self.seq_starts[seq_idx:seq_idx + 2]
    return self.hdf["targets/data/classes"][seq_t0:seq_t1]

  def get_data_dict(self, seq_idx):
    """
    :param int seq_idx:
    :rtype: dict[str,numpy.ndarray]
    """
    return {"data": self.get_data(seq_idx), "classes": self.get_targets(seq_idx)}
